/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include <editor.h>

static GtkWidget *parser_textview;
static GtkTextIter parser_iter;
static int parser_init = 0;
static char *parser_tag;

void parser_xml_parse (GtkWidget *, gpointer);
void parser_error (void *, const char *, ...);
char *parser_trim (char *);
ginfo_tips *parser_check_tips (xmlDocPtr, xmlNodePtr, int *);
ginfo_docs *parser_check_docs (xmlDocPtr, xmlNodePtr, int *);
parser_info *parser_check_info (xmlDocPtr, xmlNodePtr);
void parser_check_parser_rec (xmlDocPtr, xmlNodePtr, parser_info **,
			      parser_info **, parser_font *);
char *parser_trim_info (char *);
int parser_check_tag (xmlDocPtr, xmlNodePtr, char *, parser_font *);
void parser_parser_ok (void);

/* Add parser to notebook page */
int
parser_notebook (GtkWidget * vbox)
{
  GtkWidget *button;
  GtkWidget *sw;
  GtkTextBuffer *buffer;

  button = gtk_button_new_with_label ("Parse the XML module");
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, TRUE, 0);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);

  parser_textview = gtk_text_view_new ();
  gtk_widget_show (parser_textview);
  gtk_container_add (GTK_CONTAINER (sw), parser_textview);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (parser_textview), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (parser_textview), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (parser_textview),
			       GTK_WRAP_WORD);

  g_signal_connect ((gpointer) button, "clicked",
		    G_CALLBACK (parser_xml_parse), NULL);

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (parser_textview));
  gtk_text_buffer_create_tag (buffer, "error", "foreground", "red", NULL);
  gtk_text_buffer_create_tag (buffer, "ok", "foreground", "blue",
			      "justification", GTK_JUSTIFY_CENTER, NULL);

  return 0;
}

/* Parse the sourcebuffer */
void
parser_xml_parse (GtkWidget * w, gpointer dummy)
{
  GtkTextIter start, end;
  xmlDocPtr doc;
  xmlNodePtr cur;
  char *mem;
  char *def = NULL;
  char *name = NULL;
  char *genre = NULL;
  char *words = NULL;
  char **words_list = NULL;
  ginfo_tips *tips = NULL;
  ginfo_docs *docs = NULL;
  int docs_n = 0;
  int tips_n = 0;
  int genre_id = 0;

  gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (sourcebuffer), &start,
				      0);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (sourcebuffer), &end);

  mem =
    gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (sourcebuffer), &start, &end,
			       TRUE);

  parser_init = 0;

  parser_tag = "error";
  xmlSetGenericErrorFunc (xmlGenericErrorContext, parser_error);

  if (!(doc = xmlParseMemory (mem, strlen (mem))))
    return;

  parser_tag = NULL;
  xmlGenericError (xmlGenericErrorContext, "Parser ok!\n");

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      xmlFreeDoc (doc);
      return;
    }

  if (xmlStrcmp (cur->name, "ginfo"))
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext,
		       "RootElement must by \"ginfo\"!\n");
      xmlFreeDoc (doc);
      return;
    }

  parser_tag = NULL;
  xmlGenericError (xmlGenericErrorContext, "RootElement ok!\n");

  for (cur = cur->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "module"))
	{
	  def = xmlGetProp (cur, "default");

	  if (!(name = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*name)
	    {

	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "No name set!\n");
	      xmlFreeDoc (doc);
	      return;
	    }

	  if (def)
	    def = parser_trim (def);
	  name = parser_trim (name);

	}

      else if (!xmlStrcmp (cur->name, "genre"))
	{
	  if (!(genre = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*genre)
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "No genre set!\n");
	      xmlFreeDoc (doc);
	      return;
	    }

	  genre = parser_trim (genre);
	}

      else if (!xmlStrcmp (cur->name, "words"))
	{
	  if (!(words = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*words)
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "No words set!\n");
	      xmlFreeDoc (doc);
	      return;
	    }

	  words = parser_trim (words);
	}

      else if (!xmlStrcmp (cur->name, "tooltips"))
	{
	  if (!(tips = parser_check_tips (doc, cur, &tips_n)))
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "No tip element!\n");
	      xmlFreeDoc (doc);
	      return;
	    }
	}

      else if (!xmlStrcmp (cur->name, "docs"))
	{
	  if (!(docs = parser_check_docs (doc, cur, &docs_n)))
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "No doc element!\n");
	      xmlFreeDoc (doc);
	      return;
	    }
	}
    }

  if (!name)
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "No name set!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module name: %s\n", name);

  if (def)
    xmlGenericError (xmlGenericErrorContext, "Set flag default: YES\n");
  else
    xmlGenericError (xmlGenericErrorContext, "Set flag default: NO\n");

  if (!genre)
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "No genre set!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module genre: %s\n", genre);

  if (genre && !(genre_id = ginfo_check_genre (genre)))
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "No compatible genre!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module genre Id: %d\n", genre_id);

  if (!words)
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "no words set!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module words: %s\n", words);

  if (!(words_list = ginfo_check_words (words)))
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "no compatible words!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module words list: ok!\n");

  if (!docs)
    {
      parser_tag = "error";
      xmlGenericError (xmlGenericErrorContext, "no docs!\n");
      xmlFreeDoc (doc);
      return;
    }

  xmlGenericError (xmlGenericErrorContext, "Module docs list: ok!\n");

  if (tips)
    xmlGenericError (xmlGenericErrorContext, "Module tips list: ok!\n");
  else
    xmlGenericError (xmlGenericErrorContext, "Module tips list: no.\n");

  parser_parser_ok ();
  viewer_create_tree (docs, docs_n);
  tips_create_list (tips, tips_n);
}

/* Write in parser textview the error */
void
parser_error (void *fl, const char *str, ...)
{
  va_list va;
  char buf[1024];
  int len;
  GtkTextBuffer *buffer;
  GtkTextIter end;

  va_start (va, str);

  len = vsnprintf (buf, 1024, str, va);

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (parser_textview));

  if (!parser_init)
    {
      parser_init++;
      gtk_text_buffer_get_iter_at_offset (buffer, &parser_iter, 0);
      gtk_text_buffer_get_end_iter (buffer, &end);

      gtk_text_buffer_delete (buffer, &parser_iter, &end);

      gtk_text_buffer_get_iter_at_offset (buffer, &parser_iter, 0);
    }

  if (parser_tag)
    gtk_text_buffer_insert_with_tags_by_name (buffer, &parser_iter, buf, len,
					      parser_tag, NULL);
  else
    gtk_text_buffer_insert (buffer, &parser_iter, buf, len);

  va_end (va);
}

/* Trim the tag */
char *
parser_trim (char *tmp)
{
  int i = 0;
  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    tmp++;

  i = strlen (tmp);
  i--;

  while (tmp[i] == ' ' || tmp[i] == '\t' || tmp[i] == '\r' || tmp[i] == '\n')
    i--;

  tmp[i + 1] = 0;

  return tmp;
}

/* Check tips tags */
ginfo_tips *
parser_check_tips (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur;
  char *tip;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "tip"))
	{
	  if (!(tip = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*tip)
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "no Tip set!\n");
	      return NULL;
	    }

	  tip = parser_trim (tip);

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  ginfo_easy_tips_add (t, tip, NULL, NULL, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_tips_close (t, n);
}

/* Check the docs tags */
ginfo_docs *
parser_check_docs (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur, cur_c;
  char *chapter;
  parser_info *info;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "doc"))
	{
	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  chapter = NULL;
	  info = NULL;
	  for (cur_c = cur->children; cur_c != NULL; cur_c = cur_c->next)
	    {

	      if (!xmlStrcmp (cur_c->name, "chapter"))
		{
		  if (!
		      (chapter =
		       xmlNodeListGetString (doc, cur_c->xmlChildrenNode, 1))
		      || !*chapter)
		    {
		      ginfo_easy_docs_free (t);

		      parser_tag = "error";
		      xmlGenericError (xmlGenericErrorContext,
				       "no chapter set!\n");

		      return NULL;
		    }

		  chapter = parser_trim (chapter);

		}
	      else if (!xmlStrcmp (cur_c->name, "info"))
		{
		  if (!(info = parser_check_info (doc, cur_c->children)))
		    {
		      parser_tag = "error";
		      xmlGenericError (xmlGenericErrorContext,
				       "no info text set!\n");
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }
		}
	    }
	  if (!chapter)
	    {
	      parser_tag = "error";
	      xmlGenericError (xmlGenericErrorContext, "no chapter set!\n");
	      ginfo_easy_docs_free (t);
	      return NULL;
	    }

	  ginfo_easy_docs_add (t, chapter, NULL, (void *) info, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_docs_close (t, n);
}

/* check info tags. This functions run only parser_check_parser_rec.
 * the paeser_check_info is a recursive function.
 */
parser_info *
parser_check_info (xmlDocPtr doc, xmlNodePtr cur)
{
  parser_info *t = NULL;
  parser_info *h = NULL;
  parser_font font;

  memset (&font, 0x0, sizeof (font));
  font.scale = PANGO_SCALE_MEDIUM;

  parser_check_parser_rec (doc, cur, &t, &h, &font);

  return h;
}

/* The recursive function for info tags */
void
parser_check_parser_rec (xmlDocPtr doc, xmlNodePtr cur, parser_info ** t,
			 parser_info ** h, parser_font * font)
{

  parser_info *c;

  while (cur)
    {

      c = NULL;

      if (cur->type == XML_TEXT_NODE)
	{
	  if (!(c = (parser_info *) malloc (sizeof (parser_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;

	  c->type = INFO_TEXT;
	  c->text = xmlNodeGetContent (cur);
	  if (c->text)
	    c->text = parser_trim_info (c->text);

	  if (!c->text)
	    {
	      free (c);
	      c = NULL;
	    }
	}
      else if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!(c = (parser_info *) malloc (sizeof (parser_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->text = NULL;

	  if (!strcmp (cur->name, "nl"))
	    c->type = INFO_NEWLINE;

	  else if (!strcmp (cur->name, "space"))
	    c->type = INFO_SPACE;

	  else if (!strcmp (cur->name, "image"))
	    {
	      c->type = INFO_IMAGE;

	      if (!
		  (c->image =
		   xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
		  || !*c->image)
		{
		  free (c);
		  c = NULL;
		}
	    }

	  else
	    {
	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (parser_font));

	      if (parser_check_tag (doc, cur, (char *) cur->name, &c->font))
		{
		  free (c);
		  c = NULL;
		}
	    }
	}

      if (c)
	{
	  if (!*h)
	    *h = c;
	  else
	    (*t)->next = c;

	  *t = c;

	  if (c->type == INFO_FONT)
	    {
	      parser_check_parser_rec (doc, cur->children, t, h, &c->font);
	      if (!(c = (parser_info *) malloc (sizeof (parser_info))))
		fatal ("Error: memory.");

	      c->next = NULL;
	      c->text = NULL;
	      (*t)->next = c;
	      *t = c;

	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (parser_font));
	    }
	}

      cur = cur->next;
    }
}

/* Search the correct tags */
int
parser_check_tag (xmlDocPtr doc, xmlNodePtr cur, char *tag,
		  parser_font * font)
{
  /* STYLE */
  if (!strcmp (tag, "oblique"))
    {
      font->style = PANGO_STYLE_OBLIQUE;
      return 0;
    }

  if (!strcmp (tag, "italic"))
    {
      font->style = PANGO_STYLE_ITALIC;
      return 0;
    }

  /* VARIANT */
  if (!strcmp (tag, "small_caps"))
    {
      font->variant = PANGO_VARIANT_SMALL_CAPS;
      return 0;
    }

  /* WEIGHT */
  if (!strcmp (tag, "ultralight"))
    {
      font->weight = PANGO_WEIGHT_ULTRALIGHT;
      return 0;
    }

  if (!strcmp (tag, "light"))
    {
      font->weight = PANGO_WEIGHT_LIGHT;
      return 0;
    }

  if (!strcmp (tag, "bold"))
    {
      font->weight = PANGO_WEIGHT_BOLD;
      return 0;
    }

  if (!strcmp (tag, "ultrabold"))
    {
      font->weight = PANGO_WEIGHT_ULTRABOLD;
      return 0;
    }

  if (!strcmp (tag, "heavy"))
    {
      font->weight = PANGO_WEIGHT_HEAVY;
      return 0;
    }

  /* STRETCH */
  if (!strcmp (tag, "ultracondensed"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "extracondensed"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "condensed"))
    {
      font->stretch = PANGO_STRETCH_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semicondensed"))
    {
      font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semiexpanded"))
    {
      font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "expanded"))
    {
      font->stretch = PANGO_STRETCH_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "extraexpanded"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "ultraexpanded"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
      return 0;
    }

  /* SCALE */
  if (!strcmp (tag, "xx_small"))
    {
      font->scale = PANGO_SCALE_XX_SMALL;
      return 0;
    }

  if (!strcmp (tag, "x_small"))
    {
      font->scale = PANGO_SCALE_X_SMALL;
      return 0;
    }

  if (!strcmp (tag, "small"))
    {
      font->scale = PANGO_SCALE_SMALL;
      return 0;
    }

  if (!strcmp (tag, "large"))
    {
      font->scale = PANGO_SCALE_LARGE;
      return 0;
    }

  if (!strcmp (tag, "x_large"))
    {
      font->scale = PANGO_SCALE_X_LARGE;
      return 0;
    }

  if (!strcmp (tag, "xx_large"))
    {
      font->scale = PANGO_SCALE_XX_LARGE;
      return 0;
    }

  /* SIZE */
  if (!strcmp (tag, "size"))
    {
      char *size;
      size = xmlGetProp (cur, "value");
      if (!size || !*size)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      font->size = atoi (size) * PANGO_SCALE;

      return 0;
    }

  /* COLOR */
  if (!strcmp (tag, "foreground"))
    {
      font->foreground = xmlGetProp (cur, "value");
      if (!font->foreground || !*font->foreground)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "background"))
    {
      font->background = xmlGetProp (cur, "value");
      if (!font->background || !*font->background)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  /* JUSTIFY */
  if (!strcmp (tag, "justify"))
    {
      char *justify;
      justify = xmlGetProp (cur, "value");

      if (!justify || !*justify)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      if (!strcmp (justify, "center"))
	font->justify = GTK_JUSTIFY_CENTER;
      else if (!strcmp (justify, "left"))
	font->justify = GTK_JUSTIFY_LEFT;
      else if (!strcmp (justify, "right"))
	font->justify = GTK_JUSTIFY_RIGHT;
      else if (!strcmp (justify, "fill"))
	font->justify = GTK_JUSTIFY_FILL;
      else
	{
	  xmlGenericError (xmlGenericErrorContext,
			   "tag %s value=center|left|right\n", tag);
	  return 1;
	}

      return 0;
    }

  /* UNDERLINE */
  if (!strcmp (tag, "underline"))
    {
      font->underline = 1;
      return 0;
    }

  if (!strcmp (tag, "double_underline"))
    {
      font->underline = 2;
      return 0;
    }

  /* FACE */
  if (!strcmp (tag, "face"))
    {
      font->face = xmlGetProp (cur, "value");
      if (!font->face || !*font->face)
	{
	  xmlGenericError (xmlGenericErrorContext, "tag %s value=???\n", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "font"))
    {
      char *tmp;
      tmp = xmlGetProp (cur, "style");

      /* Style */
      if (tmp)
	{
	  if (!strcmp (tmp, "oblique"))
	    font->style = PANGO_STYLE_OBLIQUE;
	  else if (!strcmp (tmp, "italic"))
	    font->style = PANGO_STYLE_ITALIC;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Style axpected oblique|italic");
	      return 1;
	    }
	}

      /* Variant */
      tmp = xmlGetProp (cur, "variant");
      if (tmp)
	{
	  if (!strcmp (tmp, "small_cups"))
	    font->variant = PANGO_VARIANT_SMALL_CAPS;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Variant axpected small_caps");
	      return 1;
	    }
	}

      /* Weight */
      tmp = xmlGetProp (cur, "weight");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultralight"))
	    font->weight = PANGO_WEIGHT_ULTRALIGHT;
	  else if (!strcmp (tmp, "light"))
	    font->weight = PANGO_WEIGHT_LIGHT;
	  else if (!strcmp (tmp, "bold"))
	    font->weight = PANGO_WEIGHT_BOLD;
	  else if (!strcmp (tmp, "ultrabold"))
	    font->weight = PANGO_WEIGHT_ULTRABOLD;
	  else if (!strcmp (tmp, "heavy"))
	    font->weight = PANGO_WEIGHT_HEAVY;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Weight axpected ultralight|light|bold|ultrabold|heavy");
	      return 1;
	    }
	}

      /* Stretch */
      tmp = xmlGetProp (cur, "stretch");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultracondensed"))
	    font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
	  else if (!strcmp (tmp, "extracondensed"))
	    font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
	  else if (!strcmp (tmp, "condensed"))
	    font->stretch = PANGO_STRETCH_CONDENSED;
	  else if (!strcmp (tmp, "semicondensed"))
	    font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
	  else if (!strcmp (tmp, "semiexpanded"))
	    font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
	  else if (!strcmp (tmp, "expanded"))
	    font->stretch = PANGO_STRETCH_EXPANDED;
	  else if (!strcmp (tmp, "extraexpanded"))
	    font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
	  else if (!strcmp (tmp, "ultraexpanded"))
	    font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Stretch axpected ultracondensed|extracondensed|condensed|semicondensed|semiexpanded|expanded|extraexpanded|ultraexpanded");
	      return 1;
	    }
	}

      /* Scale */
      tmp = xmlGetProp (cur, "scale");
      if (tmp)
	{
	  if (!strcmp (tmp, "xx_small"))
	    font->scale = PANGO_SCALE_XX_SMALL;
	  if (!strcmp (tmp, "x_small"))
	    font->scale = PANGO_SCALE_X_SMALL;
	  if (!strcmp (tmp, "small"))
	    font->scale = PANGO_SCALE_SMALL;
	  if (!strcmp (tmp, "large"))
	    font->scale = PANGO_SCALE_LARGE;
	  if (!strcmp (tmp, "x_large"))
	    font->scale = PANGO_SCALE_X_LARGE;
	  if (!strcmp (tmp, "xx_large"))
	    font->scale = PANGO_SCALE_XX_LARGE;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Scale axpected xx_small|x_small|small|large|x_lage|xx_large");
	      return 1;
	    }
	}

      /* Size */
      tmp = xmlGetProp (cur, "size");
      if (tmp)
	{
	  font->size = atoi (tmp) * PANGO_SCALE;
	}

      /* Color */
      font->foreground = xmlGetProp (cur, "foreground");
      font->background = xmlGetProp (cur, "background");

      /* Justify */
      tmp = xmlGetProp (cur, "justify");
      if (tmp)
	{
	  if (!strcmp (tmp, "center"))
	    font->justify = GTK_JUSTIFY_CENTER;
	  else if (!strcmp (tmp, "left"))
	    font->justify = GTK_JUSTIFY_LEFT;
	  else if (!strcmp (tmp, "right"))
	    font->justify = GTK_JUSTIFY_RIGHT;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Justify axpected center|left|right");
	      return 1;
	    }
	}

      /* Underline */
      tmp = xmlGetProp (cur, "underline");
      if (tmp)
	{
	  if (!strcmp (tmp, "single"))
	    font->underline = 1;
	  else if (!strcmp (tmp, "double"))
	    font->underline = 2;
	  else
	    {
	      xmlGenericError (xmlGenericErrorContext,
			       "Font:Underline axpected single|double");
	      return 1;
	    }
	}

      /* Face */
      font->face = xmlGetProp (cur, "face");

      return 0;
    }

  xmlGenericError (xmlGenericErrorContext, "Unknow tag %s!\n", tag);

  return 1;
}

/* trim the info text */
char *
parser_trim_info (char *a)
{
  register int i = 0, j = 0;
  char *b;
  int k = 0, len;

  if (!a || !*a)
    return NULL;

  len = strlen (a);

  if (!(b = (char *) malloc (sizeof (char) * (len + 1))))
    fatal ("Error: memory.");

  while (i < len)
    {
      if (a[i] == ' ' || a[i] == '\t' || a[i] == '\r' || a[i] == '\n')
	{
	  if (!k)
	    {
	      b[j++] = ' ';
	      k = 1;
	    }
	}
      else
	{
	  b[j++] = a[i];
	  k = 0;
	}

      i++;
    }

  b[j] = 0;

  for (i = k = 0; i < j; i++)
    if (b[i] != ' ')
      k++;

  if (!k)
    {
      free (b);
      return NULL;
    }

  return b;
}

/* Write the correct message */
#define PARSE_OK "*** Parse Ok! ***\n**** Look the viewer and the tips pages. ***"
void
parser_parser_ok (void)
{
  GtkTextBuffer *buffer;

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (parser_textview));
  gtk_text_buffer_insert_with_tags_by_name (buffer, &parser_iter, PARSE_OK,
					    strlen (PARSE_OK), "ok", NULL);
}

/* EOF */
